/*
  Copyright (c) 1998 - 2008
  ILK  -  Tilburg University
  CNTS -  University of Antwerp
 
  This file is part of Timbl

  Timbl is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.

  Timbl is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, see <http://www.gnu.org/licenses/>.

  For questions and suggestions, see:
      http://ilk.uvt.nl/software.html
  or send mail to:
      Timbl@uvt.nl
*/


#include <string>
#include <map>
#include <fstream>
#include <sstream>
#include <iomanip>
#include <cassert>

#include "timbl/MsgClass.h"
#include "timbl/Common.h"
#include "timbl/StringOps.h"
#include "timbl/Types.h"
#include "timbl/Options.h"
#include "timbl/Tree.h"
#include "timbl/Instance.h"
#include "timbl/Statistics.h"
#include "timbl/neighborSet.h"
#include "timbl/BestArray.h"
#include "timbl/IBtree.h"

#ifdef USE_LOGSTREAMS
#include "timbl/LogStream.h"
#else
typedef std::ostream LogStream;
#define Log(X) (X)
#define Dbg(X) (X)
#endif

#include "timbl/MBLClass.h"
#include "timbl/TimblExperiment.h"

namespace Timbl {
  using namespace std;

  void IG_Experiment::InitInstanceBase(){
    srand( RandomSeed() );
    default_order();
    set_order();
    runningPhase = TrainWords;
    InstanceBase = new IG_InstanceBase( EffectiveFeatures(), 
					ibCount,
					(RandomSeed()>=0), 
					false, KeepDistributions() );
  }
  
  void IG_Experiment::show_results( ostream& outfile, 
				    const string& dString,
				    const TargetValue *Best,
				    const double Distance ) {
    show_org_input( outfile );
    outfile << CodeToStr(Best->Name());
    if ( Verbosity(DISTRIB) ){
      outfile << " " << dString;
    }
    if ( Verbosity(DISTANCE) ) {
      int OldPrec = outfile.precision(DBL_DIG-1);
      outfile.setf(ios::showpoint);
      outfile.width(8);
      outfile << " " << Distance;
      outfile.precision(OldPrec);
    }
    if ( Verbosity(MATCH_DEPTH) ){
      outfile << " " << last_depth << ":" << (last_leaf?"L":"N");
    }
    outfile << endl;
    showBestNeighbors( outfile );
  }
  
  void IG_Experiment::initExperiment( bool ){ 
    if ( !ExpInvalid() ) {
      if ( !MBL_init ){  // do this only when necessary
	stats.clear();
	delete confusionInfo;
	confusionInfo = 0;
	if ( Verbosity(ADVANCED_STATS) )
	  confusionInfo = new ConfusionMatrix( Targets->ValuesArray.size() );
	if ( !is_copy ){
	  InitWeights();
	  if ( do_diversify )
	    diverseWeights();
	  srand( random_seed );
	}
	MBL_init = true;
      }
    }
  }  

  bool IG_Experiment::checkFile( const string& fileName ){
    if ( !TimblExperiment::checkFile( fileName ) )
      return false;
    else if ( IBStatus() != Pruned ){
      Warning( "you tried to apply the IGTree algorithm on a complete,"
	       "(non-pruned) Instance Base" );
      return false;
    }
    else {
      if ( num_of_neighbors != 1 ){
	Warning( "number of neighbors set to 1 for IGTree test!" );
	num_of_neighbors = 1;
      }
      if ( decay_flag  != Zero ){
	Warning( "Decay impossible for IGTree test, (k=1)" );
	decay_flag = Zero;
      }
      if ( GlobalMetric != Overlap ){
	Warning( "Metric set to Overlap for IGTree test." );
	GlobalMetric = Overlap;
      }
    }
    return true;
  }
    
  bool IG_Experiment::build_file_index( const string& file_name, 
					featureMultiIndex& fmIndex ){
    bool result = true;
    string Buffer;
    stats.clear();
    size_t cur_pos = 0;
    // Open the file.
    //
    ifstream datafile( file_name.c_str(), ios::in);
    if ( InputFormat() == ARFF )
      skipARFFHeader( datafile );
    cur_pos = datafile.tellg();
    if ( !nextLine( datafile, Buffer ) ){
      Error( "cannot start learning from in: " + file_name );
      result = false;    // No more input
    }
    else if ( !chopLine( Buffer ) ){
      Error( "no useful data in: " + file_name );
      result = false;    // No more input
    }
    else {
      if ( !Verbosity(SILENT) ) {
	Info( "Phase 2: Building index on Datafile: " + file_name );
	time_stamp( "Start:     ", 0 );
      }
      bool found;
      bool go_on = true;
      while( go_on ){ 
	// The next Instance to store. 
	chopped_to_instance( TrainWords );
	FeatureValue *fv0 = CurrInst.FV[0];
	FeatureValue *fv1 = CurrInst.FV[1];
	pair<FeatureValue*,streamsize> fsPair = make_pair( fv1, cur_pos );
	featureMultiIndex::iterator it = fmIndex.find( fv0 );
	if ( it != fmIndex.end() )
	  it->second.insert( fsPair );
	else {
	  MultiIndex mi;
	  mi.insert( fsPair );
	  fmIndex[fv0] = mi;
	}
	if ((stats.dataLines() % Progress() ) == 0) 
	  time_stamp( "Indexing:  ", stats.dataLines() );
	found = false;
	while ( !found && 
		( cur_pos = datafile.tellg(),
		  nextLine( datafile, Buffer ) ) ){
	  found = chopLine( Buffer );
	  if ( !found ){
	    Warning( "datafile, skipped line #" + 
		     toString<int>( stats.totalLines() ) +
		     "\n" + Buffer );
	  }
	}
	go_on = found;
      }
      time_stamp( "Finished:  ", stats.dataLines() );
    }
    return result;
  }

  bool IG_Experiment::Learn( const string& FileName ){
    bool result = true;
    if ( ExpInvalid() ||
	 !ConfirmOptions() ){
      result = false;
  }
    else if ( CurrentDataFile == "" ){
      if ( FileName == "" ){
	Warning( "unable to build an InstanceBase: No datafile defined yet" );
	result = false;
      }
      else {
	if ( !Prepare( FileName ) || ExpInvalid() ){
	  result = false;
	}
      }
    }
    else if ( FileName != "" &&
	      CurrentDataFile != FileName ){
      Error( "Unable to Learn from file '" + FileName + "'\n"
	     "while previously instantiated from file '" + 
	     CurrentDataFile + "'" );
      result = false;
    }
    if ( result ) {
      InitInstanceBase();
      featureMultiIndex fmIndex;
      result = build_file_index( CurrentDataFile, fmIndex );
      if ( result ){
	stats.clear();
	if ( !Verbosity(SILENT) ) {
	  Info( "\nPhase 3: Learning from Datafile: " + CurrentDataFile );
	  time_stamp( "Start:     ", 0 );
	}
	string Buffer;
	IG_InstanceBase *PartInstanceBase = 0;
	TargetValue *TopTarget = Targets->MajorityClass();
	// Open the file.
	//
	ifstream datafile( CurrentDataFile.c_str(), ios::in);
	//
	Feature *Feat =  Features[Permutation[0]];
	IVCmaptype::const_iterator it = Feat->ValuesMap.begin();
	//
	// don't ask why. but we need to iterate the Map here, not the Array
	//
	while ( it !=  Feat->ValuesMap.end() ){
	  FeatureValue *the_fv = (FeatureValue*)(it->second);
	  featureMultiIndex::const_iterator fmIt = fmIndex.find( the_fv );
	  if ( fmIt == fmIndex.end() ){
	    FatalError( "panic" );
	  }
	  if ( igOffset() > 0 && fmIt->second.size() > igOffset() ){
	    IVCmaptype::const_iterator it2
	      = Features[Permutation[1]]->ValuesMap.begin();
	    IG_InstanceBase *TmpInstanceBase = 0;
	    TmpInstanceBase = new IG_InstanceBase( EffectiveFeatures(), 
						   ibCount,
						   (RandomSeed()>=0), 
						   false, 
						   true );
	    while ( it2 != Features[Permutation[1]]->ValuesMap.end() ){
	      FeatureValue *the2fv = (FeatureValue*)(it2->second);
	      typedef MultiIndex::const_iterator mit;
	      pair<mit,mit> b = fmIt->second.equal_range( the2fv );
	      for ( mit i = b.first; i != b.second; ++i ){
		datafile.seekg( i->second );
		nextLine( datafile, Buffer );
		chopLine( Buffer );
		// Progress update.
		//
		if (( stats.dataLines() % Progress() ) == 0) 
		  time_stamp( "Learning:  ", stats.dataLines() );
		chopped_to_instance( TrainWords );
		if ( !PartInstanceBase ){
		  PartInstanceBase = new IG_InstanceBase( EffectiveFeatures(), 
							  ibCount,
							  (RandomSeed()>=0), 
							  false, 
							  true );
		}
		PartInstanceBase->AddInstance( CurrInst );
	      }
	      if ( PartInstanceBase ){
// 		Info( "finished handling secondary feature '"
// 		      + the2fv->Name() + "'" );
//		time_stamp( "Start Pruning:    " );
		PartInstanceBase->Prune( TopTarget, 2 );
//		time_stamp( "Finished Pruning: " );
		if ( !TmpInstanceBase->Merge( PartInstanceBase, the_fv ) ){
		  FatalError( "Merging InstanceBases failed. PANIC" );
		  return false;
		}
		delete PartInstanceBase;
		PartInstanceBase = 0;
	      }
	      ++it2;
	    }
//	    time_stamp( "Start Final Pruning:    " );
	    TmpInstanceBase->specialPrune( TopTarget );
//	    time_stamp( "Finished Final Pruning: " );
	    if ( !InstanceBase->Merge( TmpInstanceBase, the_fv ) ){
	      FatalError( "Merging InstanceBases failed. PANIC" );
	      return false;
	    }
	    delete TmpInstanceBase;
	  }
	  else {
	    MultiIndex::const_iterator mIt = fmIt->second.begin();
	    while ( mIt != fmIt->second.end() ){
	      datafile.seekg( mIt->second );
	      nextLine( datafile, Buffer );
	      chopLine( Buffer );
	      // Progress update.
	      //
	      if (( stats.dataLines() % Progress() ) == 0) 
		time_stamp( "Learning:  ", stats.dataLines() );
	      chopped_to_instance( TrainWords );
	      if ( !PartInstanceBase )
		PartInstanceBase = new IG_InstanceBase( EffectiveFeatures(), 
							ibCount,
							(RandomSeed()>=0), 
							false, 
							true );
	      PartInstanceBase->AddInstance( CurrInst );
	      ++mIt;
	    }
	    if ( PartInstanceBase ){
// 	      Info( "1 finished handling feature '" + the_fv->Name() + "'" );
//	      time_stamp( "Start Pruning:    " );
	      PartInstanceBase->Prune( TopTarget );
	      //	      time_stamp( "Finished Pruning: " );
	      if ( !InstanceBase->Merge( PartInstanceBase, the_fv ) ){
		FatalError( "Merging InstanceBases failed. PANIC" );
		return false;
	      }
	      delete PartInstanceBase;
	      PartInstanceBase = 0;
	    }
	  }
	  ++it;
	}
	time_stamp( "Finished:  ", stats.dataLines() );
	if ( !Verbosity(SILENT) )
	  IBInfo( *Log(mylog) );
      }
    }
    return result;
  }

  bool IG_Experiment::checkLine( const string& line ){
    if ( !TimblExperiment::checkLine( line ) )
      return false;
    else if ( IBStatus() != Pruned ){
      Warning( "you tried to apply the IGTree algorithm on a complete,"
	       "(non-pruned) Instance Base" );
      return false;
    }
    else {
      if ( num_of_neighbors != 1 ){
	Warning( "number of neighbors set to 1 for IGTree test!" );
	num_of_neighbors = 1;
      }
      if ( decay_flag != Zero ){
	Warning( "Decay impossible for IGTree test, (k=1)" );
	decay_flag = Zero;
      }
      if ( GlobalMetric != Overlap ){
	Warning( "Metric set to Overlap for IGTree test." );
	GlobalMetric = Overlap;
      }
    }
    return true;
  }      
  
  const TargetValue *IG_Experiment::LocalClassify( const Instance& Inst,
						   double& Distance, 
						   bool& exact ){
    last_depth = -1;
    last_leaf = false;
    exact = false;
    bool Tie = false;
    initExperiment();
    bestResult.reset( beamSize, normalisation, norm_factor, Targets );
    const TargetValue *TV = NULL;
    const ValueDistribution *ResultDist 
      = InstanceBase->IG_test( Inst, last_depth, last_leaf, TV );
    if ( last_depth == 0 ){
      // when level 0, ResultDist == TopDistribution
      TV = InstanceBase->TopTarget( Tie );
    }
    Distance = sum_remaining_weights( last_depth );
    if ( InstanceBase->PersistentD() && ResultDist ){
      if ( last_depth == 0 )
	bestResult.addTop( ResultDist );
      else
	bestResult.addConstant( ResultDist );
    }
    if ( confusionInfo )
      confusionInfo->Increment( Inst.TV, TV );
    bool correct = Inst.TV && ( TV == Inst.TV );
    if ( correct ){
      stats.addCorrect();
      if ( Tie )
	stats.addTieCorrect();
    }
    else if ( Tie )
      stats.addTieFailure();
    return TV;
  }

  void IG_Experiment::testing_info( ostream& os,
				    const string& FileName,
				    const string& OutFile ){
    if ( Verbosity(OPTIONS) )
      ShowSettings( os );
    os << endl << "Starting to test, Testfile: " << FileName << endl
       << "Writing output in:          " << OutFile << endl
       << "Algorithm     : IGTree" << endl;
    show_ignore_info( os );
    show_weight_info( os );
    os << endl;
  }

  bool IG_Experiment::WriteInstanceBase( const string& FileName ){
    bool result = false;
    if ( ConfirmOptions() ){
      ofstream outfile( FileName.c_str(), ios::out | ios::trunc );
      if (!outfile) {
	Warning( "can't open outputfile: " + FileName );
      }
      else {
	if ( !Verbosity(SILENT) )
	  Info( "Writing Instance-Base in: " + FileName );
	if ( PutInstanceBase( outfile ) ){
	  string tmp = FileName;
	  tmp += ".wgt";
	  ofstream wf( tmp.c_str() );
	  if ( !wf ){
	    Error( "can't write default weightfile " + tmp );
	    result = false;
	  }
	  else if ( !writeWeights( wf ) )
	    result = false;
	  else if ( !Verbosity(SILENT) )
	    Info( "Saving Weights in " + tmp );
	  result = true;
	}
      }
    }
    return result;
  }
  
  bool IG_Experiment::GetInstanceBase( istream& is ){
    bool result = false;
    bool Pruned;
    bool Hashed;
    int Version;
    string range_buf;
    if ( !get_IB_Info( is, Pruned, Version, Hashed, range_buf ) ){
      Error( "Can'n retrieve Instance-Base\n" );
    }
    else if ( !Pruned ){
      Error( "Instance-base is NOT Pruned!, invalid for " +
	     toString(algorithm) + " Algorithm" );
    }
    else {
      TreeOrder = DataFile;
      Initialize();
      if ( !get_ranges( range_buf ) ){
	Warning( "couldn't retrieve ranges..." );
      }
      else {
	srand( RandomSeed() );
	InstanceBase = new IG_InstanceBase( EffectiveFeatures(), 
					    ibCount,
					    (RandomSeed()>=0), 
					    Pruned,
					    KeepDistributions() );
	int pos=0;
	for ( size_t i=0; i < NumOfFeatures(); ++i ){
	  Features[i]->SetWeight( 1.0 );
	  if ( Features[Permutation[i]]->Ignore() )
	    PermFeatures[i] = NULL;
	  else 
	    PermFeatures[pos++] = Features[Permutation[i]];
	}
	if ( Hashed )
	  result = InstanceBase->ReadIB( is, PermFeatures,
					 Targets, 
					 TargetStrings, FeatureStrings,
					 Version ); 
	else
	  result = InstanceBase->ReadIB( is, PermFeatures, 
					 Targets, 
					 Version ); 
	if ( result ){
	  if ( !InstanceBase->HasDistributions() ){
	    if ( Verbosity(DISTRIB) ){
	      Info( "Instance base doesn't contain Distributions, "
		    "+vDB option disabled ...."  );
	      ResetVerbosityFlag(DISTRIB);
	    }
	    KeepDistributions( false );
	  }
	}
      }
    }
    return result;
  }

  bool IG_Experiment::ReadInstanceBase( const string& FileName ){
    bool result = false;
    if ( ConfirmOptions() ){
      ifstream infile( FileName.c_str(), ios::in );
      if ( !infile ) {
	Error( "can't open: " + FileName );
      }
      else {
	if ( !Verbosity(SILENT) )
	  Info( "Reading Instance-Base from: " + FileName );
	if ( GetInstanceBase( infile ) ){
	  if ( !Verbosity(SILENT) ){
	    writePermutation( my_log() );
	  }
	  string tmp = FileName;
	  tmp += ".wgt";
	  ifstream wf( tmp.c_str() );
	  if ( !wf ){
	    Error( "cant't find default weightsfile " + tmp );
	  }
	  else if ( readWeights( wf, CurrentWeighting() ) ){
	    WFileName = tmp;
	    if ( !Verbosity(SILENT) ){
	      Info( "Reading weights from " + tmp );
	    }
	  }
	  result = true;
	}
      }
    }
    return result;
  }
      
}
